import subprocess as proc


class BlastDB(object):
    makeblastdb_cmd = (
        'makeblastdb -in {sequence_path} -input_type {input_type} '
        '-dbtype {dbtype} -title {title} -out {db_path}')
    
    def __init__(
        self, sequence_path:str, title:str, db_path:str,
        input_type:str='fasta', dbtype:str='nucl',
        run=True
    ):
        """Creates a new blast database.
        
        Parameters
        ----------
        sequence_path : str
            Path to sequences file.
        title : str
            Title of the database
        db_path : str
            Path where to save the database files
        input_type : str, optional
            Type of sequences input file, by default 'fasta'
        dbtype : str, optional
            Type of database to be constructed, by default 'nucl'
        
        Raises
        ------
        SystemError
            Raised when makeblastdb returns a non-zero return code.
        """
        self.title = title
        self.db_path = db_path
        self.dbtype = dbtype
        
        if run:
            self.run(sequence_path, input_type)

    def run(self, sequence_path, input_type):
        makeblastdb_cmd = type(self).makeblastdb_cmd.format(
            sequence_path=sequence_path,
            input_type=input_type,
            dbtype=self.dbtype,
            title=self.title,
            db_path=self.db_path)
        res = proc.run(makeblastdb_cmd, shell=True)
        if res.returncode != 0:
            raise SystemError(f'makeblastdb failed: {makeblastdb_cmd}')

    def __repr__(self):
        return f'BlastDB({self.title}, {self.db_path}, dbtype={self.dbtype})'



import pandas as pd


class BlastSeq(object):
    def __init__(self, name:str, metadata:dict=None):
        """Creates a new sequence representation.

        Parameters
        ----------
        name : str
            Unique name of the sequence
        metadata : dict, optional
            Relavant data about the sequence, by default None
        """
        self._name = name
        self.metadata = metadata
        self._query_matches = {}
    
    @property
    def name(self):
        return self._name

    @property
    def query_matches(self):
        return self._query_matches    
    
    @property
    def num_query_matches(self):
        return len(self._query_matches)
    
    def reciprocal_matches(self) -> list:
        """Returns the list of reciprocal blastn matches involving
        this sequence.

        Returns
        -------
        list
            List of ReciprocalBlastMatch objects representing
            reciprocal matches between this sequence and zero
            or more other sequences.
        """
        matches = []
        # look into query matches 
        for query, meta in self._query_matches.items():
            for r, rmeta in query._query_matches.items():
                if r == self:
                    matches.append(
                        ReciprocalBlastMatch(self, query, meta, rmeta))
        return matches
    
    def add_query_match(self, query, metadata=None):
        """Adds a blast match between this sequence as reference and
        the other sequence as query.

        Parameters
        ----------
        query : BlastSeq
            BlastSeq object representating the query sequence.
        metadata : dict-like, optional
            Relevant data about the match. 
            Usually a row in the blast results table, by default None
        """
        self._query_matches[query] = BlastMatch(query, self, metadata)
        
    def __repr__(self):
        return f'BlastSeq({self.name})'

    def __eq__(self, other):
        return (
            (self.name == other.name) and 
            len(self.query_matches) == len(other.query_matches)
        )
    
    def __hash__(self):
        return hash(self.name)


class BlastMatch(object):
    def __init__(self, query:BlastSeq, ref:BlastSeq, metadata=None):
        """Create a representation of a blast match between two sequences.

        Parameters
        ----------
        query : BlastSeq
            Query sequence
        ref : BlastSeq
            Reference/subject (database) sequence
        metadata : dict-like, optional
            Relevant metadata about the match.
            Usually a row in the blast results table, by default None
        """
        self.query = query
        self.ref = ref
        # metadata
        self.length = None
        self.mismatch = None
        self.gapopen = None
        self.qstart = None
        self.qend = None
        self.sstart = None
        self.send = None
        self.evalue = None
        self.bitscore = None
        self.qlen = None
        self.slen = None
        self.qcovs = None
        self.sstrand = None
        
        self._parse_metadata_(metadata)

    @property
    def metadata(self):
        return pd.Series({
            'pident': self.pident,
            'length': self.length,
            'mismatch': self.mismatch,
            'gapopen': self.gapopen,
            'qstart': self.qstart,
            'qend': self.qend,
            'sstart': self.sstart,
            'send': self.send,
            'evalue': self.evalue,
            'bitscore': self.bitscore,
            'qlen': self.qlen,
            'slen': self.slen,
            'qcovs': self.qcovs,
            'sstrand': self.sstrand,
        })

    def _parse_metadata_(self, metadata):
        if metadata is not None:
            self.pident = (
                metadata['pident'] if 'pident' in metadata.keys() else None)
            self.length = (
                metadata['length'] if 'length' in metadata.keys() else None)
            self.mismatch = (
                metadata['mismatch'] if 'mismatch' in metadata.keys() else None)
            self.gapopen = (
                metadata['gapopen'] if 'gapopen' in metadata.keys() else None)
            self.qstart = (
                metadata['qstart'] if 'qstart' in metadata.keys() else None)
            self.qend = metadata['qend'] if 'qend' in metadata.keys() else None
            self.sstart = (
                metadata['sstart'] if 'sstart' in metadata.keys() else None)
            self.send = metadata['send'] if 'send' in metadata.keys() else None
            self.evalue = (
                metadata['evalue'] if 'evalue' in metadata.keys() else None)
            self.bitscore = (
                metadata['bitscore'] if 'bitscore' in metadata.keys() else None)
            self.qlen = metadata['qlen'] if 'qlen' in metadata.keys() else None
            self.slen = metadata['slen'] if 'slen' in metadata.keys() else None
            self.qcovs = (
                metadata['qcovs'] if 'qcovs' in metadata.keys() else None)
            self.sstrand = (
                metadata['sstrand'] if 'sstrand' in metadata.keys() else None)
    
    def __repr__(self):
        return f'BlastMatch(query={self.query.name}, ref={self.ref.name})'

    def __eq__(self, other):
        return (
            (self.query == other.query) and
            (self.ref == other.ref) and
            (self.length == other.length) and
            (self.mismatch == other.mismatch) and
            (self.gapopen == other.gapopen) and
            (self.qstart == other.qstart) and
            (self.qend == other.qend) and
            (self.sstart == other.sstart) and
            (self.send == other.send) and
            (self.evalue == other.evalue) and
            (self.bitscore == other.bitscore) and
            (self.qlen == other.qlen) and
            (self.slen == other.slen)
        )


class ReciprocalBlastMatch(object):
    tuple_index = pd.MultiIndex.from_tuples([
        (match, prop) 
        for match in ['forward', 'reverse']
        for prop in [
            'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend',
            'sstart', 'send', 'evalue', 'bitscore', 'qlen', 'slen', 'qcovs',
            'sstrand']
        ], names=['match', 'property'])

    def __init__(
        self, seq1:BlastSeq, seq2:BlastSeq, forward_match:BlastMatch, reverse_match:BlastMatch
    ):
        """Create a representation of a reciprocal blast match between
        two sequences.

        Parameters
        ----------
        seq1 : BlastSeq
            Query sequence
        seq2 : BlastSeq
            Reference/subject (database) sequence
        forward_match : BlastMatch
            Match object for the forward (seq1 v. seq2) blast search.
        reverse_match : BlastMatch
            Match object for the reverse (seq21 v. seq1) blast search.
        """
        self.seq1 = seq1
        self.seq2 = seq2
        # metadata
        self.match1 = forward_match
        self.match2 = reverse_match
    
    @property
    def metadata(self):
        values = (
            list(self.match1.metadata.values) + 
            list(self.match2.metadata.values))
        return pd.Series(values, index=type(self).tuple_index)

    def __repr__(self):
        return (
            f'ReciprocalBlastMatch(seq1={self.seq1.name}, '
            f'seq2={self.seq2.name})'
        )

    def __eq__(self, other):
        return (
            (
                (self.seq1 == other.seq1) or
                (self.seq1 == other.seq2)
            ) and
            (
                (self.seq2 == other.seq2) or
                (self.seq2 == other.seq1)
            ) and
            ((
                (self.match1 == other.match1) and
                (self.match2 == other.match2)
            ) or 
            (
                (self.match1 == other.match2) and
                (self.match2 == other.match1)
            ))
        )

    def __hash__(self):
        return hash(tuple(sorted([self.seq1, self.seq2], key=lambda s: s.name)))



from reciprocalblast.db import BlastDB
from reciprocalblast.search import (
    BlastnSearch,
    ReciprocalBlastnSearch
)
from reciprocalblast.match import (
    BlastSeq,
    BlastMatch,
    ReciprocalBlastMatch,
)

__version__ = '0.1.0'



import subprocess as proc

import pandas as pd

from reciprocalblast.db import BlastDB
from reciprocalblast.match import BlastSeq


class BlastnSearch(object):
    blastn_cmd = (
        'blastn -query {query} -out {out} -db {db} '
        '-task {task} -outfmt {outfmt} '
        '-max_target_seqs {max_target_seqs} -max_hsps {max_hsps} '
        '-num_threads {num_threads}')
    
    def __init__(
        self, query_path:str, out_path:str, db:BlastDB,
        task:str='blastn', outfmt:str='"6 std qlen slen qcovs sstrand"',
        max_target_seqs:int=1, max_hsps:int=1, 
        eval_threshold:float=1e-10, num_threads:int=1, run:bool=True,
    ):
        """Performs a blastn search of query sequences against a blast database.

        Parameters
        ----------
        query_path : str
            Path of the query sequences file.
        out_path : str
            Path to save the blastn results table.
        db : BlastDB
            BlastDB object that represents the blast database to be used.
        task : str, optional
            blastn search algorithm to used, by default 'blastn'
        outfmt : str, optional
            Output format configuration, 
            by default '"6 std qlen slen qcovs sstrand"'
        max_target_seqs : int, optional
            Maximum number of aligned sequences to keep, by default 1
        max_hsps : int, optional
            Maximum number of high-scoring pair (HSP) to keep for each query-subject alignment, by default 1
        eval_threshold : float, optional
            Maximum expect value for retaining hits, by default 1e-10
        num_threads : int, optional
            Number of threads to use in the blast search, by default 1
        run : bool, optional
            If True, runs the blast search and overwrites existing results.
            Otherwise, assumes that blast search has been performed previously
            and will only import the results, by default True
        """
        self.query_path = query_path
        self.out_path = out_path
        self.db = db.db_path if isinstance(db, BlastDB) else db
        self.task = task
        self.outfmt = outfmt
        self.max_target_seqs = max_target_seqs
        self.max_hsps = max_hsps
        self.result = None
        
        if run:
            self.run(num_threads=num_threads)
        self.result = self.parse_result(out_path, eval_threshold)

    def run(self, num_threads:int=1):
        """Runs blast search.

        Parameters
        ----------
        num_threads : int, optional
            Number of threads to use in the blast search, by default 1

        Raises
        ------
        SystemError
            Raised when blastn returns a non-zero return code.
        """
        blastn_cmd = type(self).blastn_cmd.format(
            query=self.query_path,
            out=self.out_path,
            db=self.db,
            task=self.task,
            outfmt=self.outfmt,
            max_target_seqs=self.max_target_seqs,
            max_hsps=self.max_hsps,
            num_threads=num_threads)
        res = proc.run(blastn_cmd, shell=True)
        if res.returncode != 0:
            raise SystemError(f'blastn failed: {blastn_cmd}')

    @staticmethod
    def parse_result(path:str, eval_threshold:float=1e-10) -> pd.DataFrame:
        """Reads the blast result csv into a pandas DataFrame.
        
        Parameters
        ----------
        path : str
            Path of blastn results file.
        eval_threshold : float, optional
            Maximum expect value for retaining hits, by default 1e-10
        
        Returns
        -------
        pandas.DataFrame
            Returns the blastn results after removing hits above the
            specified expect value threshold.
        """
        col_labels = ['qaccver', 'saccver', 'pident', 'length', 
                      'mismatch', 'gapopen', 'qstart', 'qend', 
                      'sstart', 'send', 'evalue', 'bitscore', 
                      'qlen', 'slen', 'qcovs', 'sstrand']
        ref_df = pd.read_csv(path, sep='\t', header=None, index_col=None)
        ref_df.columns = col_labels
        filtered_df = ref_df[ref_df['evalue'] < eval_threshold]
        return filtered_df.set_index(['qaccver', 'saccver'])
    
    def __repr__(self):
        return f'BlastnSearch(query={self.query_path}, db={self.db}, task={self.task})'


class ReciprocalBlastnSearch(object):
    def __init__(
        self, seq1_fasta:str, seq2_fasta:str, seq1_db:BlastDB, seq2_db:BlastDB,
        seq1_out:str='', seq2_out:str='',
        task:str='blastn', outfmt:str='"6 std qlen slen qcovs sstrand"',
        max_target_seqs:int=1, max_hsps:int=1, 
        eval_threshold:int=1e-10, num_threads:int=1,
        run:bool=True,
    ):
        """Performs reciprocal blastn search between two sets of sequences.
        
        Parameters
        ----------
        seq1_fasta : str
            Path to file with the first set of sequences.
        seq2_fasta : str
            Path to file with the second set of sequences.
        seq1_db : BlastDB
            Path to blast db created from the first set of sequences.
        seq2_db : BlastDB
            Path to blast db created from the second set of sequences.
        seq1_out : str, optional
            Path where to save the forward (seq1 v. seq2) blastn search,
            by default ''
        seq2_out : str, optional
            Path where to save the reverse (seq2 v. seq1) blastn search,
            by default ''
        task : str, optional
            blastn search algorithm to used, by default 'blastn'
        outfmt : str, optional
            Output format configuration, 
            by default '"6 std qlen slen qcovs sstrand"'
        max_target_seqs : int, optional
            Maximum number of aligned sequences to keep, by default 1
        max_hsps : int, optional
            Maximum number of high-scoring pair (HSP) to keep for each query-subject alignment, by default 1
        eval_threshold : float, optional
            Maximum expect value for retaining hits, by default 1e-10
        num_threads : int, optional
            Number of threads to use in the blast search, by default 1
        run : bool, optional
            If True, runs the blast search and overwrites existing results.
            Otherwise, assumes that blast search has been performed previously
            and will only import the results, by default True
        """
        self.seq1_fasta = seq1_fasta
        self.seq2_fasta = seq2_fasta
        self.seq1_db = (
            seq1_db.db_path if isinstance(seq1_db, BlastDB) else seq1_db)
        self.seq2_db = (
            seq2_db.db_path if isinstance(seq2_db, BlastDB) else seq2_db)
        self.seq1_out = seq1_out if seq1_out else seq1_fasta + '.blastn.txt'
        self.seq2_out = seq2_out if seq2_out else seq2_fasta + '.blastn.txt'
        self.task = task
        self.outfmt = outfmt
        self.max_target_seqs = max_target_seqs
        self.max_hsps = max_hsps
        self.result1 = None
        self.result2 = None
        self._reciprocal_matches = None
        
        if run:
            self.run(num_threads=num_threads)

        self.result1 = BlastnSearch.parse_result(self.seq1_out, eval_threshold)
        self.result2 = BlastnSearch.parse_result(self.seq2_out, eval_threshold)

    def run(self, num_threads=1):
        """Runs reciprocal blast search.
        
        Parameters
        ----------
        num_threads : int, optional
            Number of threads to use in the blast search, by default 1
        
        Raises
        ------
        SystemError
            Raised when either forward or reverse blastn search
            returns a non-zero return code.
        """
        res1 = proc.run(
            BlastnSearch.blastn_cmd.format(
                query=self.seq1_fasta,
                out=self.seq1_out,
                db=self.seq2_db,
                task=self.task,
                outfmt=self.outfmt,
                max_target_seqs=self.max_target_seqs,
                max_hsps=self.max_hsps,
                num_threads=num_threads,
            ),
            shell=True,
        )
        if res1.returncode != 0:
            raise SystemError('Forward blastn search failed to run.')
        res2 = proc.run(
            BlastnSearch.blastn_cmd.format(
                query=self.seq2_fasta,
                out=self.seq2_out,
                db=self.seq1_db,
                task=self.task,
                outfmt=self.outfmt,
                max_target_seqs=self.max_target_seqs,
                max_hsps=self.max_hsps,
                num_threads=num_threads,
            ),
            shell=True,
        )
        if res2.returncode != 0:
            raise SystemError('Reverse blastn search failed to run.')
            
    @staticmethod
    def parse_result(out1:str, out2:str, eval_threshold=1e-10) -> pd.DataFrame:
        """Reads the blast result csv into a pandas DataFrame.
        
        Parameters
        ----------
        out1 : str
            Path of forward blastn results file.
        out2 : str
            Path of reverse blastn results file.
        eval_threshold : float, optional
            Maximum expect value for retaining hits, by default 1e-10
        
        Returns
        -------
        pandas.DataFrame
            Returns the blastn results after removing hits above the
            specified expect value threshold.
        """
        result1 = BlastnSearch.parse_result(out1, eval_threshold)
        result2 = BlastnSearch.parse_result(out2, eval_threshold)
        # Create relationship graphs
        blastseq_d = {}
        # Process result 1
        for k, v in result1.iterrows():
            # process keys
            query_id, ref_id = k            
            # query id
            if query_id not in blastseq_d.keys():
                query_blastseq = BlastSeq(query_id)
                blastseq_d[query_id] = query_blastseq
            else:
                query_blastseq = blastseq_d[query_id]
            # ref id
            if ref_id not in blastseq_d.keys():
                ref_blastseq = BlastSeq(ref_id)
                blastseq_d[ref_id] = ref_blastseq
            else:
                ref_blastseq = blastseq_d[ref_id]
            ref_blastseq.add_query_match(query_blastseq, v)
        # Process result 2
        for k, v in result2.iterrows():
            # process keys
            query_id, ref_id = k
            # query id
            if query_id not in blastseq_d.keys():
                query_blastseq = BlastSeq(query_id)
                blastseq_d[query_id] = query_blastseq
            else:
                query_blastseq = blastseq_d[query_id]
            # ref id
            if ref_id not in blastseq_d.keys():
                ref_blastseq = BlastSeq(ref_id)
                blastseq_d[ref_id] = ref_blastseq
            else:
                ref_blastseq = blastseq_d[ref_id]
            ref_blastseq.add_query_match(query_blastseq, v)
        # Get reciprocal matches
        reciprocal_matches = []
        for seq in blastseq_d.values():
            reciprocal_matches += seq.reciprocal_matches()
        return list(set(sorted(
            reciprocal_matches, 
            key=lambda s: s.seq1.name + s.seq2.name
        )))

    def reciprocal_matches(self, eval_threshold:float=1e-10) -> list:
        """Returns the list of reciprocal blast matches computed from
        the forward and reverse blastn searches.
        
        Parameters
        ----------
        eval_threshold : float, optional
            Maximum expect value for retaining hits, by default 1e-10
        
        Returns
        -------
        list
            List of ReciprocalBlastMatch objects representing
            reciprocal matches between two sequences.
        """
        if not self._reciprocal_matches:
            self._reciprocal_matches =\
                self.parse_result(self.seq1_out, self.seq2_out, eval_threshold)
        return self._reciprocal_matches

    def reciprocal_matched_df(self, eval_threshold:float=1e-10) -> pd.DataFrame:
        """Returns the list of reciprocal blast matches as a pandas DataFrame.
        
        Parameters
        ----------
        eval_threshold : float, optional
            Maximum expect value for retaining hits, by default 1e-10
        
        Returns
        -------
        list
            DataFrame containing the reciprocal matches between two sequences.
        """
        index = pd.MultiIndex.from_tuples(
            [(m.seq1.name, m.seq2.name) for m in self.reciprocal_matches()],
            names=['seq1', 'seq2'])
        return pd.DataFrame.from_records(
            [m.metadata for m in reciprocal_matches], 
            index=index)

    def __repr__(self):
        return (
            f'ReciprocalBlastnSearch(seq1_fasta={self.seq1_fasta}, '
            f'seq2_fasta={self.seq2_fasta}, task={self.task})')



from setuptools import setup, find_packages
from reciprocalblast import __version__

setup(
    name='reciprocalblast',
    version=__version__,
    author='Kent Kawashima',
    author_email='kentkawashima@gmail.com',
    description='Perform reciprocal blastn analysis',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    keywords=['blast', 'search', 'sequence', 'dna', 'bioinformatics'],
    packages=find_packages(),
)
